/**
 * @file	eth_queue.c
 * @author	cd185990
 * @date	2010/07/10
 * @brief	PriorityQueue Implementation File
 *
 * Last Change Base: 		$LastChangedRevision: 456 $
 * Last Changed By: 		$LastChangedBy: cd185990 $
 * Last Changed On: 		$LastChangedDate: 2014-03-03 15:35:00 +0100 (Mon, 03 Mar 2014) $
 *
 * Copyright: 2010 - 2014, Mobile Network Performance Group, RWTH Aachen
 */

#include "stdbool.h"
//#include "stdio.h"

#include "eth_queue.h"


// Hide internal handling of the queue
typedef struct eq_t {
	/// Number of priorities beside top priority
	unsigned char eq_priorities;
	/// Number of slots for top priority queue
	unsigned char eq_top_length;
	/// Number of total slots available for lower priority queues (equally distributed)
	unsigned char eq_low_length;
	/// Array holding the start/stop IDs of ring buffer
	unsigned char eq_id_array[(EQ_MAX_LENGTH-EQ_RESERVED)*2];
	/// Indication whether partial priority queue is full
	bool eq_is_priority_full[(EQ_MAX_LENGTH-EQ_RESERVED)];
	/// Array holding the length values of the information stored in the buffer
	unsigned short int eq_lengths[(EQ_MAX_LENGTH-EQ_RESERVED)];
	/// Enable overwrite bit
	bool eq_enable_overwrite;
	/// Status initialized
	bool eq_initialized;
} eq;

eq queue;


signed char eq_init(unsigned char priorities, unsigned char top_length, bool enable_overwrite) {
	queue.eq_priorities = priorities;
	queue.eq_top_length = top_length;
	queue.eq_low_length = priorities == 0 ? 0 : (unsigned char)((EQ_MAX_LENGTH-EQ_RESERVED - queue.eq_top_length) / queue.eq_priorities);
	queue.eq_enable_overwrite = enable_overwrite;

	if ( queue.eq_priorities > (EQ_MAX_LENGTH-EQ_RESERVED-1) ||
		(queue.eq_low_length == 0 && queue.eq_priorities != 0) ||
		 queue.eq_low_length > EQ_MAX_LENGTH ||
		 top_length == 0 ||
		 top_length > (EQ_MAX_LENGTH-EQ_RESERVED)) {
//		xil_printf("Software Ethernet Queue: Wrong parameter selection.\n");
		return CALLBACK_ERROR;
	}

	// resetting all values
	unsigned char i;
	for (i = 0; i < 2*(EQ_MAX_LENGTH-EQ_RESERVED); ++i)
		queue.eq_id_array[i] = 0;
	for (i = 0; i < (EQ_MAX_LENGTH-EQ_RESERVED); ++i)
		queue.eq_lengths[i] = 0;
	for (i = 0; i <= priorities; ++i)
		queue.eq_is_priority_full[i] = false;

	// initializing the ring buffer start/stop values
	for (i = 2; i < 2+queue.eq_priorities*2; ++i) { // top priority queue has IDs {0,1}
		if (even(i))
			queue.eq_id_array[i] = queue.eq_top_length+(i-2)/2*queue.eq_low_length; // start id
		else
			queue.eq_id_array[i] = queue.eq_top_length+(i-3)/2*queue.eq_low_length; // stop id
	}

	queue.eq_initialized = true;
	return CALLBACK_OK;
}


signed char eq_enqueue(eq_result *result, unsigned char priority, unsigned short int length) {
	result->buffer = EQ_MAX_LENGTH;
	result->length = 0;
	if (queue.eq_initialized == false || priority > queue.eq_priorities) return CALLBACK_ERROR;

	// if (is queue ful
	if (queue.eq_is_priority_full[priority] == true) {
		if (queue.eq_enable_overwrite) {
			result->buffer = queue.eq_id_array[2*priority]; // give back the soon to be overwritten value
			result->length = length;
			queue.eq_lengths[result->buffer] = length;
			queue.eq_id_array[2*priority] = ++queue.eq_id_array[2*priority];
			if (priority == 0) {
				if ((queue.eq_id_array[0] % queue.eq_top_length) == 0)
					queue.eq_id_array[0] = 0;
			} else {
				if (((queue.eq_id_array[2*priority]-queue.eq_top_length-(priority-1)*queue.eq_low_length) % (queue.eq_low_length)) == 0)
					queue.eq_id_array[2*priority] = queue.eq_top_length + (priority-1) * queue.eq_low_length;
			}
			queue.eq_id_array[2*priority+1] = queue.eq_id_array[2*priority];
		}
		return CALLBACK_ERROR;
	} else { // else(is queue full)
		result->buffer = queue.eq_id_array[2*priority+1];
		result->length = length;
		queue.eq_lengths[result->buffer] = length;
		queue.eq_id_array[2*priority+1] = ++queue.eq_id_array[2*priority+1];
		if (priority == 0) {
			if ((queue.eq_id_array[1] % queue.eq_top_length) == 0)
				queue.eq_id_array[1] = 0;
		} else {
			if (((queue.eq_id_array[2*priority+1]-queue.eq_top_length-(priority-1)*queue.eq_low_length) % (queue.eq_low_length)) == 0)
				queue.eq_id_array[2*priority+1] = queue.eq_top_length + (priority-1) * queue.eq_low_length;
		}
		if (queue.eq_id_array[2*priority] == queue.eq_id_array[2*priority+1])
			queue.eq_is_priority_full[priority] = true;
		return CALLBACK_OK;
	} // end(is queue full)
}


signed char eq_dequeue(eq_result *result) {
	result->buffer = EQ_MAX_LENGTH;
	result->length = 0;
	if (queue.eq_initialized == false || eq_queue_empty(-1)) return CALLBACK_ERROR;

	unsigned char i;
	for (i = 0; i <= queue.eq_priorities; ++i) {
		if (!((queue.eq_id_array[2*i] == queue.eq_id_array[2*i+1]) && !queue.eq_is_priority_full[i])) {
			if (queue.eq_is_priority_full[i])
				queue.eq_is_priority_full[i] = false;
			result->buffer = queue.eq_id_array[2*i];
			result->length = queue.eq_lengths[result->buffer];
			queue.eq_id_array[2*i] = ++queue.eq_id_array[2*i];
			if (i == 0) {
				if ((queue.eq_id_array[0] % queue.eq_top_length) == 0)
					queue.eq_id_array[0] = 0;
			} else {
				if (((queue.eq_id_array[2*i]-queue.eq_top_length-(i-1)*queue.eq_low_length) % (queue.eq_low_length)) == 0)
					queue.eq_id_array[2*i] = queue.eq_top_length + (i-1) * queue.eq_low_length;
			}
			return CALLBACK_OK;
		}
	}

	return CALLBACK_ERROR;
}


bool eq_queue_empty(signed char priority) {
	if (queue.eq_initialized == false || priority > (EQ_MAX_LENGTH-EQ_RESERVED)) return true; // is there a better return value?

	if (priority < 0) {
		bool result = true;
		unsigned char i;
		for (i = 0; i <= queue.eq_priorities; ++i) {
			if (!result) break;
			result &= (queue.eq_id_array[2*i] == queue.eq_id_array[2*i+1]) && !queue.eq_is_priority_full[i];
		}
		return result;
	} else
		return (queue.eq_id_array[2*priority] == queue.eq_id_array[2*priority+1]) && !queue.eq_is_priority_full[priority];
}


bool eq_queue_full(signed char priority) {
	if (priority > (EQ_MAX_LENGTH-EQ_RESERVED)) return false; // is there a better return value?

	if (priority < 0) {
		bool result = false;
		unsigned char i = 0;
		while (!result && (i < (EQ_MAX_LENGTH-EQ_RESERVED))) {
			result |= queue.eq_is_priority_full[i];
			i++;
		}
		return result;
	} else
		return queue.eq_is_priority_full[priority];
}


signed char eq_front(eq_result *result) {
	result->buffer = EQ_MAX_LENGTH;
	result->length = 0;
	if (queue.eq_initialized == false || eq_queue_empty(-1)) return CALLBACK_ERROR;

	unsigned char i;
	for (i = 0; i <= queue.eq_priorities; ++i) {
		if (!((queue.eq_id_array[2*i] == queue.eq_id_array[2*i+1]) && !queue.eq_is_priority_full[i])) {
			result->buffer = queue.eq_id_array[2*i];
			result->length = queue.eq_lengths[result->buffer];
			return CALLBACK_OK;
		}
	}

	return CALLBACK_ERROR;
}
