/**
 * @file	measurement.h
 * @author	cd185990
 * @date	2013/05/05
 * @brief	Measurement Interface
 *
 * Last Change Base: 		$LastChangedRevision: 475 $
 * Last Changed By: 		$LastChangedBy: cd185990 $
 * Last Changed On: 		$LastChangedDate: 2014-07-09 16:48:53 +0200 (Wed, 09 Jul 2014) $
 *
 * Copyright: 2010 - 2014, Mobile Network Performance Group, RWTH Aachen
 */

#ifndef MEASUREMENT_H_
#define MEASUREMENT_H_

#include "xparameters.h"
#include "stdbool.h"

#include "mnp.h"
#include "sync.h" // mandatory requirement
#include "warpmac.h"
#include "warpphy.h"

#define MNP_MEAS_VERSION			"0.6.13"

#define MNP_MEAS_TIMER_ID			3
#define MNP_MEAS_TXBUFFER_ID1		EQ_RESID_MEAS1 // finish packet start index (aggregated values, +X for station X)
#define MNP_MEAS_TXBUFFER_ID2		EQ_RESID_MEAS2 // status packet
#define MNP_MEAS_MAXSTATIONS		(1+5) // maximum no. of stations that have own result data structure
#define MNP_MEAS_SNR_BATCHES		210 // 0dBm -> -105dBm in 0.5dB steps
#define MNP_MEAS_PCKSRC				1
#define MNP_MEAS_PCKDST				2
#define MNP_MEAS_PCKDFIN_LENGTH		(sizeof(measPacketDataFinish)-sizeof(syncPacketHeader))///sizeof(unsigned int))
#define MNP_MEAS_COUNTERNR			6

typedef enum {
	PACKET_MEAS = 50,
	PACKET_MEAS_END,
	PACKET_MEAS_UNSYNC
} PacketMeasurement;

typedef enum {
	MPDataStart = MNP_SYNC_PACKETTYPES,
	MPDataStop,
	MPDataStatus,
	MPDataFinish,
	MPBackStart,
	MPBackStop,
	MPRejoin,
	MPSNRStart,
	MPSNRStop,
	MPSNRStatus,
	MPSNRFinish,
	MPSNRFinishSrc
} measPacketType;

typedef enum {
	RCToff,
	RCTrec, // all recovery types accumulated
	RCTrecUni,
	RCTrecBi,
	RCTrecFault,
//	RCTrecFail, // recorded as RCToff (compare to stateTransitions[3] and stateTransitions[4])
	RCTall // all offline and recovery types accumulated
} rotationCounterType;

/////////////////////////////////////
/// Data Measurement in TokenRing ///
/////////////////////////////////////
typedef struct {
	syncPacketHeader subheader;
	unsigned char source;
	unsigned char destination;
	unsigned short int deadline; // 4 Byte full
	unsigned int packetNumber; // 4 Byte full
	unsigned short int packetLength;
	unsigned short int periodicity; // 4 Byte full
	unsigned int time; // 4 Byte full
	unsigned char power;
	unsigned char reserved0;
	unsigned short int reserved1; // 4 Byte full
} measPacketDataStart, measDataSetup; // packet transmitted over ETH

typedef struct {
	syncPacketHeader subheader;
	unsigned int value;
} measPacketDataStatus; // packets transmitted over ETH

typedef struct { // MNP_MEAS_PCKDFIN_LENGTH = ... *4Byte full
	syncPacketHeader subheader;
	unsigned int dataPackets[4]; // [0] packetNumber, [1] packetsOkay, [2] packetsFailed, [3] badPayload
	unsigned int rcvdPackets[5]; // [0] only original, [1] original and relay, [2] original and re-tx, [3] only relay, [4] only re-tx
	unsigned int lostDataPackets;
	unsigned int dataPacketProb[2]; // [0] offlineEnqueuedPackets, [1] dropped from ETH queue
	unsigned int stateTransitions[5]; // [0] M>R, [1] I>O (leave), [2] I>O (txError), [3] R>O (failed recovery), [4] I>O (aborted recovery)
	unsigned int rotations[MNP_MEAS_COUNTERNR][27]; // [0]-[20] direct counter, [21] more than 20 rotations, [22] more than 1e2, [23] more than 1e3, [24] more than 1e4, [25] more than 1e5, [26] more than 1e6
	unsigned int recoveryTypes[3]; // [0] one-direction error, [1] both-direction error, [2] bridging gap of disappeared station
	unsigned int totalRotations;
	unsigned int sentData[2]; // [0] data TX without relay set, [1] data TX with relay set
	unsigned int sentToken[8]; // lastState w/o relay set [0] SOLICITING, [1] SENDING, [2] JOINING, [3] IDLE, w/ relay set: [4] SOLICITING, [5] SENDING, [6] JOINING, [7] IDLE
	unsigned int relayedPackets[2]; // [0] TOKEN, [1] DATA
	unsigned int rcvdToken[5]; // [0] only original, [1] original and relay, [2] original and re-tx, [3] only relay, [4] only re-tx
	unsigned int selfRing;
	unsigned int claimTokenOffs[4]; // [0] off after normalCT, [1] off after patchedCT, [2] off after leaveCT, [3] bypassed Recovery after CT normal
	unsigned int scrambledPackets[3]; // [0] unidentified packet, [1] identified packet in wrong state, [2] received packet while "sleeping"
	unsigned int delayProfile[47]; // [0]-[39] 0..1,1..2,...,39..40ms, [40] <50, [41] <60, [42] <70, [43] <80, [44] <90, [45] <100, [46] >100
	unsigned int evaluatedStation;
	unsigned int successors[MNP_MEAS_MAXSTATIONS]; // [0]-[MAXSTATIONS]: successor occurrences
	unsigned int predecessors[MNP_MEAS_MAXSTATIONS]; // [0]-[MAXSTATIONS]: predecessor occurrences
	unsigned int rings[MNP_MEAS_MAXSTATIONS]; // [0]-[MAXSTATIONS]: ring occurrences
	unsigned int awareness[2*9]; // second dimension: [0] = not late, [1] late
} measPacketDataFinish, measDataResults __attribute__((__aligned__(8))); // packets transmitted over ETH

typedef struct {
	unsigned long long timestamp; // 8 byte full
	unsigned char packetType;
	unsigned char reserved0;
	unsigned short int reserved1;
	unsigned int reserved2; // 8 byte full
} packetDataMeasurement; // packet transmitted over the WARP PHY

typedef struct {
	measDataSetup setup;
	unsigned int rotationCounter[MNP_MEAS_COUNTERNR];
	unsigned int rotationCounterSingle[MNP_MEAS_COUNTERNR];
	unsigned char unsynchronizedDest;

	unsigned int runPeriodicity;
	unsigned long long counter;
	unsigned long long runTime;
	unsigned long long runDeadline;
	unsigned short int divider;
	unsigned short int dividerCounter;
	bool finished;
	bool isDiscovered;
	unsigned short int lastAckedPck;
	unsigned char pcIPAddress;
	unsigned char pcMACAddress[6];
	unsigned char batchCount;
	short int rejoinCounter;
} measData __attribute__((__aligned__(8)));

typedef struct {
	syncPacketHeader subheader;
	unsigned char destination;
} packetBackStart, packetBackStop;

typedef struct {
	syncPacketHeader subheader;
	unsigned short int warpDelay; // ms
} measPacketRejoin;


////////////////////////////////////
/// SNR Measurement in TokenRing ///
////////////////////////////////////
typedef struct {
	syncPacketHeader subheader;
	unsigned char source;
	unsigned char testPattern1;
	unsigned char testPattern2;
	unsigned char power; // 4 Byte full
	unsigned int packetNumber; // 4 Byte full
	unsigned short int packetLength;
	unsigned short int periodicity; // 4 Byte full
} measPacketSNRStart, measSNRSetup; // packet transmitted over ETH

typedef struct {
	syncPacketHeader subheader;
	unsigned char source;
	unsigned char reserved0;
	unsigned char reserved1;
	unsigned char reserved2; // 4 Byte full
	unsigned int packetNumber; // 4 Byte full
} measPacketSNRStatus, measPacketSNRFinishSrc __attribute__((__aligned__(8))); // packet transmitted over ETH

typedef struct {
	syncPacketHeader subheader;
	unsigned char source;
	unsigned char destination;
	unsigned char batchStartNumber;
	unsigned char reserved; // 4 Byte full
	unsigned int values[70][3]; // 3 * 70
	unsigned int receivedPackets[4]; // [0] none, [1] only first, [2] only second, [3] both
} measPacketSNRFinish __attribute__((__aligned__(8))); // packet transmitted over ETH

typedef struct {
	measSNRSetup setup;
	unsigned int results[MNP_MEAS_SNR_BATCHES][3]; // [0] packetsOkay, [1] badPayload, [2] badHeader
	unsigned int transmittedPackets;
	unsigned int receivedPackets[4]; // [0] none, [1] only first, [2] only second, [3] both
	unsigned int runPeriodicity;
	bool finished;
	bool isSweaping;
} measSNR __attribute__((__aligned__(8)));

#endif /* MEASUREMENT_H_ */
