#include "warpmac.h"
#include "xparameters.h"
#include "mnp.h"
#include "warp_timer_regMacros.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>


#include "eth_queue.h"
#include "../station.h"
#include "../shared_memory.h"
#include "../helper.h"

#define ENET_LINK_SPEED 100

int warpmac_getMyId() {
	return station_id;
}

//Instantiates the Maccontrol control structure
Maccontrol controlStruct;


unsigned int csmaActive;
unsigned int last_carrierSense; // for edge detection

//Points to a user-provided Macframe where parsed header information will be copied to
Macframe* rxPacket;

//Global variable for tracking active LED outputs
unsigned int LEDHEX_Outputs;
unsigned int ledStates;
unsigned int leftHex;
unsigned int rightHex;

//"Low" LED states
unsigned int ledStatesLow[2];
unsigned char ledStatesIndexLow;
//"High" LED states
unsigned int ledStatesHigh[2];
unsigned char ledStatesIndexHigh;
//Dip-switch state
unsigned char dipswState;
//Debug GPIO state
unsigned char debugGPIO;

//Global variable to track SISO vs. MIMO mode
unsigned char warpmac_TxMultiplexingMode;

unsigned char pktGen_allow;
unsigned int dummyPacketInterval;
unsigned char startPktGeneration;
unsigned int dummyPacketLength;

//These the addresses pointed to by these variables are set by the
//user via the callback registration functions

void (*usr_upbutton) ();
void (*usr_leftbutton) ();
void (*usr_middlebutton) ();
void (*usr_rightbutton) ();
void (*usr_badHeaderCallback) ();
int (*usr_goodHeaderCallback) ();
void (*usr_goodHeaderCallbackPost) ();
void (*usr_timerCallback) ();
void (*usr_dataFromNetworkLayer) ();
void (*usr_mgmtPkt) ();
void (*usr_uartRecvCallback) ();

//Null handler, for callbacks the user doesn't re-assign
void nullCallback(void* param){};
int nullCallback_i(void* param){return 0;};

///@brief Initializes the framework and all hardware peripherals.
///
///This function sets reasonable default values for many of the parameters of the MAC, configures
///interrupts and exceptions, configures Ethernet, and finally initializes the custom peripherals
///such as the radio controller, the PHY, the packet detector, and the automatic gain control block.
void warpmac_init() {
	
	xil_printf("Initializing WARPMAC v18.0:\r\n");
	
	//XStatus Status;

	//Initialize global variables
	startPktGeneration = 0;
	dummyPacketLength = 1470;
	debugGPIO = 0;
	LEDHEX_Outputs = 0;
	ledStates = 0;
	leftHex = 0;
	rightHex = 0;
	ledStatesLow[0] = 1;
	ledStatesLow[1] = 2;
	ledStatesIndexLow = 0;
	ledStatesHigh[0] = 4;
	ledStatesHigh[1] = 8;
	ledStatesIndexHigh = 0;
	warpmac_TxMultiplexingMode = 0;
	
	csmaActive = 0;
	last_carrierSense = 0;

	//Assign a null handler to all the callbacks; the user will override these after initialization
	usr_upbutton = nullCallback;
	usr_leftbutton = nullCallback;
	usr_middlebutton = nullCallback;
	usr_rightbutton = nullCallback;
	usr_badHeaderCallback = nullCallback;
	usr_goodHeaderCallback = nullCallback_i;
	usr_goodHeaderCallbackPost = nullCallback;
	usr_timerCallback = nullCallback;
	usr_dataFromNetworkLayer = nullCallback;
	usr_mgmtPkt = nullCallback;
	usr_uartRecvCallback = nullCallback;
	
	//Initialize the PHY
	warpphy_init();

	/*********************MAC Parameters*************************/
	//Sensible defaults; user can override
	controlStruct.maxReSend = 4;
	controlStruct.currentCW = 0;
	controlStruct.maxCW = 5;

	controlStruct.pktBuf_phyRx = 0;
	controlStruct.pktBuf_emacRx = 1;
	controlStruct.dummyPacketMode = 0;
	
	//warpphy_setBuffs(controlStruct.pktBuf_emacRx, controlStruct.pktBuf_phyRx);
	
	/************************************************************/
	
	/*******************Ethernet**********************************/
	xil_printf("\tInitializing TEMAC...");
	
	xil_printf("complete! - Link Speed %d\r\n", ENET_LINK_SPEED);
	/************************************************************/
	/*
	//Manually call the UserIO ISR once to store the initial value of the buttons/switches
	//This is especially important for applications where the value of the DIP switch means something at boot
	userIO_handler((void *)&GPIO_UserIO);
	
	//Set the backoff timer to pause during medium-busy events
	// The user can either disable this with a similar call (using mode=TIMER_MODE_NOCARRIERSENSE)
	// or set the carrier sense threshold so high it that it will never see "busy"
	warp_timer_setMode(BACKOFF_TIMER, TIMER_MODE_CARRIERSENSE); 
	
	//In this design, we dedicate 1 of the 8 timers int the warp_timer core to polling user IO
	//like the uart and push buttons. The rate of this poll is controlled by USERIO_POLLRATE.*/
	/*warp_timer_setTimer(USERIO_TIMER, 0, USERIO_POLLRATE); 
	warp_timer_setMode(USERIO_TIMER, TIMER_MODE_NOCARRIERSENSE);
	warp_timer_start(USERIO_TIMER);*/
	
	return;
}

///@brief Top-level polling function
///
///This function polls each peripheral in the design (PHY, timer and Ethernet)
///User applications should call this function frequently
void warpmac_pollPeripherals() {
	//This function polls the three main peripherals: the PHY, timer and EMAC
	warpmac_pollPhy();
	warpmac_pollTimer();
	warpmac_pollDataSource();
	
	usleep(10);
	
	return;
}

///@brief Polls the OFDM PHY
///
///Function checks the status of the PHY receiver
//inline void warpmac_pollPhy() {
void warpmac_pollPhy() {
	if(*station_rx_header_indicator == RX_GOOD_HEADER)
	{
		//printf("station % d received packet: ", station_id);
		//print_packet_buffer((void *)warpphy_getBuffAddr(controlStruct.pktBuf_phyRx));

		phyRx_goodHeader_handler();

		*station_rx_header_indicator = RX_NO_HEADER;
		return;
	}

	if(*station_rx_header_indicator == RX_BAD_HEADER)
	{
		phyRx_badHeader_handler();

		*station_rx_header_indicator = RX_NO_HEADER;
		return;
	}
}

///@brief Polls the Ethernet
///
///Function checks the status of the ping and pong Emac buffers
///and calls the effective-ISR when either is filled.
inline void warpmac_pollDataSource() {

	//If the user code has disabled new data (probably becuase they're still working on the previous packet)
	// return immediately without actually polling anything. This allows the TEAMC and LLFIFO to buffer
	// incoming packets, dropping them at the wire if too many arrive before the user is ready
	if( controlStruct.enableDataFromNetwork == 0 )
		return;

	//If we're in dummyPacket mode, and it's time to generate a new packet,
	// call the user callback immediately, as if a packet had been received over the wire into pktBuf_emacRx
	if(controlStruct.dummyPacketMode && startPktGeneration && pktGen_allow)
	{
		pktGen_allow = 0;
		usr_dataFromNetworkLayer(dummyPacketLength, controlStruct.pktBuf_emacRx);
	}

	//Two ways to check if there are packets in the LLFIFO:
	// 1) Read the ISR register for XLLF_INT_RC_MASK (receive complete)
	// 2) Read the RDFO register for a count of received packets in the FIFO
	//Option 2 works well in testing
	//if(XIo_In32(EMAC_FIFO_BaseAddr+XLLF_ISR_OFFSET) & XLLF_INT_RC_MASK)


	unsigned char *buf;
	// get the user defined buffer, or in case of the EQ the temporary buffer
	buf = (unsigned char *)warpphy_getBuffAddr(controlStruct.pktBuf_emacRx)+NUM_HEADER_BYTES;

	int nread;

	struct sockaddr_in si_other;
	unsigned int size = sizeof(si_other);

	int skipDataCallback = 0;

	if((nread = recvfrom(sockfd, (void*)(buf + 42), 2048 - NUM_HEADER_BYTES - 42, 0, &si_other, &size)) > 0)
	{
		// LLFIFO and TEMAC error handling (see also warpmac_init())
		// maybe not sufficient
//		if ((XLlFifo_Status(&FifoInstance) & XLLF_INT_RXERROR_MASK) || // LLFIFO error
//		    (XLlTemac_Status(&TemacInstance) & XTE_INT_RECV_ERROR_MASK)) { // TEMAC error
//			xil_printf("Error in LLFIFO/TEMAC.\r\n");
//			return;
//		}

		//Clear the LLFIFO status bits
		//XIo_Out32(EMAC_FIFO_BaseAddr+XLLF_ISR_OFFSET, 0x0FFFFFFF);

		//Call the ethernet packet handler

		//simulate ethernet header
		int i;
		for(i=0; i<6; ++i)
		{
			buf[i] = 0x00; //dest mac addr
		}

		for(i=6; i<12; ++i)
		{
			buf[i] = 0x00; //src mac addr
		}

		buf[12] = 0x08; //type (ETH_TYPE_4)
		buf[13] = 0x00;

		//simulate ip header
		buf[14] = 0x45; //version & length (4 & 20 Byte)

		buf[15] = 0x00; //tos (0)

		buf[16] = ((20 + 8 + nread) & 0xff00) >> 8; //total length (20 + 8 + nread)
		buf[17] = (20 + 8 + nread) & 0x00ff;

		buf[18] = 0; //identification (0)
		buf[19] = 0;

		buf[20] = 0x40; //flags & fragment offset (Don't fragment & 0)
		buf[21] = 0x00;

		buf[22] = 0x40; //ttl (64)

		buf[23] = 0x11; //protocol (UDP)

		buf[24] = 0x00; //checksum (0)
		buf[25] = 0x00;

		buf[26] = si_other.sin_addr.s_addr & 0x000000ff; // src ip addr
		buf[27] = (si_other.sin_addr.s_addr & 0x0000ff00) >> 8;
		buf[28] = (si_other.sin_addr.s_addr & 0x00ff0000) >> 16;
		buf[29] = (si_other.sin_addr.s_addr & 0xff000000) >> 24;

		buf[30] = station_ip_addr.sin_addr.s_addr & 0x000000ff; // dest ip addr
		buf[31] = (station_ip_addr.sin_addr.s_addr & 0x0000ff00) >> 8;
		buf[32] = (station_ip_addr.sin_addr.s_addr & 0x00ff0000) >> 16;
		buf[33] = (station_ip_addr.sin_addr.s_addr & 0xff000000) >> 24;

		//simulate udp header
		buf[34] = (si_other.sin_port & 0x00ff); // src udp port
		buf[35] = (si_other.sin_port & 0xff00) >> 8;

		buf[36] = (station_udp_port & 0xff00) >> 8; // dest udp port
		buf[37] = (station_udp_port & 0x00ff);

		buf[38] = ((8 + nread) & 0xff00) >> 8; //length (8 + nread)
		buf[39] = (8 + nread) & 0x00ff;

		buf[40] = 0x00; //checksum (0)
		buf[41] = 0x00;

		/*printf("Ethernet packet received:\n");

		for(i=0; i<2048; ++i)
		{
			if(i%16 == 0)
			{
				printf("\n");
			}

			printf("%2x ", buf[i]);
		}*/

		//Wait for the DMA to be idle
		//warpmac_waitForDMA();

		//Set DMA to non-increment source, increment dest addresses
		//XDmaCentral_SetControl(&DmaCentralInst, XDMC_DMACR_DEST_INCR_MASK);

		//Read the Rx packet length; the packet must be read from the FIFO immediately after this read
		//RxPktLength = XIo_In32(EMAC_FIFO_BaseAddr+XLLF_RLF_OFFSET);

		//Transfer the Ethernet frame from the FIFO to the PHY buffer
		//XDmaCentral_Transfer(&DmaCentralInst,
		//					(u8 *)(EMAC_FIFO_BaseAddr+XLLF_RDFD_OFFSET),
		//					(u8 *)pktBufPtr,
		//					RxPktLength);

		// if the sync mode is enabled, other code will handle ongoing procedures, e.g. in sync.c
		//if (inSyncMode == 1) {
		//	return;
		//}

		if( (skipDataCallback == 0) && (controlStruct.dummyPacketMode == 0))	{
			//This call does *not* wait for the DMA to finish, since the FIFO/DMA will be much faster than an OFDM PHY Tx
			usr_dataFromNetworkLayer(14 + 20 + 8 + nread, controlStruct.pktBuf_emacRx);
		}
	}

	return;
}

unsigned char eth_tx_buffer[2048];

void warpmac_prepPktToNetwork(void* thePkt, unsigned int length)
{
	memcpy(eth_tx_buffer, thePkt, length);
}

void warpmac_startPktToNetwork(unsigned int length)
{
	struct sockaddr_in destaddr;

	bzero(&destaddr,sizeof(destaddr));
	destaddr.sin_family = AF_INET;
	memcpy(&(destaddr.sin_addr),(void*)(eth_tx_buffer+30),4);
	memcpy(&(destaddr.sin_port),(void*)(eth_tx_buffer+36),2);
	//destaddr.sin_addr.s_addr=inet_addr("127.0.0.1");
	//destaddr.sin_port=htons(7500);

	//printf("Sending packet to %s:%d length: %d\n\n",
			//inet_ntoa(destaddr.sin_addr), ntohs(destaddr.sin_port), length);

	int size = sizeof(destaddr);



	int sendto_return = sendto(sockfd, (void*)(eth_tx_buffer+42), length - 42, 0,(struct sockaddr *)&destaddr, size);


	//printf("send error: %d, %s\n", errno, strerror(errno));
	/*int i;
	for(i=0; i<2048; ++i)
	{
		if(i%16 == 0)
		{
			printf("\n");
		}

		printf("%2x ", eth_tx_buffer[i]);
	}*/


}

///@brief "Good Header" handler for the PHY layer.
///
///This function is the handler for the RX physical layer peripheral.
///This function is only called if the packet header has no bit errors
///(i.e., the packet header passes CRCs in the PHY). This handler will
///be asserted *before* the packet is fully received. This function assumes
///the user's callback will poll the PHY to determine if the packet is eventually
///good or bad
void phyRx_goodHeader_handler() {
	//Copy the received packet's header into the rxPacket struct
	memcpy((void *)(&(rxPacket->header)), (void *)warpphy_getBuffAddr(controlStruct.pktBuf_phyRx), (size_t)NUM_HEADER_BYTES);

	//Strip the header and checksum from the total number of bytes reported to user code
	if(rxPacket->header.length == NUM_HEADER_BYTES) {
		//Header-only packet (like an ACK) has 0 payload bytes
		rxPacket->header.length = 0;
	} else {
		//Other packets have payload bytes; subtract off header and 4-byte payload checksum lengths and (if coded) tail bytes
		rxPacket->header.length = (rxPacket->header.length) - NUM_HEADER_BYTES - NUM_PAYLOAD_CRC_BYTES - NUM_PAYLOAD_TAIL_BYTES;
	}

	//If the Rx packet is too big, ignore the payload
	// This should never happen; the Tx code will never send a pkt bigger than Ethernet MTU
	if(rxPacket->header.length>MY_XEM_MAX_FRAME_SIZE) {
		rxPacket->header.length = 0;
	}

	//Pass the received packet to the user handler
	usr_goodHeaderCallback(rxPacket);
	usr_goodHeaderCallbackPost(rxPacket); // post processing
}

///@brief "Bad Header" handler for the PHY layer.
///
///This function is the handler for the RX physical layer peripheral.
///This function is only called if the packet fails CRC. The received
///packet is passed to the user's callback, and the PHY is reset.
void phyRx_badHeader_handler() {
	//Bad packets, by definition, have no payload bytes which can be trusted
	// The user handler is called with no arguments, since there is no real packet data to process
	// If the header happend to be good but the payload bad, the user should use the good header
	// interrupt to process the header
	usr_badHeaderCallback();
}



///@brief Sets user callback function for various system events
///
///@param event ID of the event triggering the callback; one of EVENT_* values from warpmac.h
///@param handler Function pointer for user-provided callback function
void warpmac_setCallback(int event, void (*handler)()) {

	switch(event)
	{
		case EVENT_UPBUTTON:
			usr_upbutton = handler;
			break;
		case EVENT_LEFTBUTTON:
			usr_leftbutton = handler;
			break;
		case EVENT_RIGHTBUTTON:
			usr_rightbutton = handler;
			break;
		case EVENT_MIDDLEBUTTON:
			usr_middlebutton = handler;
			break;
		case EVENT_TIMER:
			usr_timerCallback = handler;
			break;
		case EVENT_DATAFROMNETWORK:
			usr_dataFromNetworkLayer = handler;
			break;
		case EVENT_MGMTPKT:
			usr_mgmtPkt = handler;
			break;
		case EVENT_PHYGOODHEADER:
			usr_goodHeaderCallback = handler;
			break;
		case EVENT_PHYBADHEADER:
			usr_badHeaderCallback = handler;
			break;
		case EVENT_UARTRX:
			usr_uartRecvCallback = handler;
			break;
		default:
			//Invalid event ID; do nothing
			break;
	}
	
	return;
}

void warpmac_pollUserIO()
{
	int nread;
	char	rbuf[2048];

	if ((nread = read(station_pty_fd, rbuf, 512)) < 0)
	{
		//xil_printf("Sim error: read error from stdin");
	}

	if(nread > 0)
	{
		rbuf[nread] = 0;
		//xil_printf("UART: Received %d bytes: %s\n", nread, buf);

		int i;
		for(i=0; i < nread; ++i)
		{
			usr_uartRecvCallback(rbuf[i]);
		}
	}

	if(*station_left_button)
	{
		usr_leftbutton();
		*station_left_button = 0;
	}

	if(*station_right_button)
	{
		usr_rightbutton();
		*station_right_button = 0;
	}

	if(*station_up_button)
	{
		usr_upbutton();
		*station_up_button = 0;
	}

	if(*station_middle_button)
	{
		usr_middlebutton();
		*station_middle_button = 0;
	}
}



void warpmac_pollTimer() 
{
	long current_time = get_time_nsec();

	if(/*csmaActive == 1 && simTimers[0].mode == TIMER_MODE_CARRIERSENSE
				&& */last_carrierSense == 0 && *carrier_sense == 1) //rising edge
	{
		warp_timer_pause(0);

		Marker m;
		m.station = warpmac_getMyId();
		m.marker_type = MARKER_CSMA_RISING_EDGE;
		m.time = get_time_nsec();
		m.added_to_list = 0;
		add_marker(&m, warpmac_getMyId());

		//printf("b");
	}

	int i;
	for(i=0; i<NUM_SIM_TIMERS; ++i)
	{
		if(simTimers[i].isActive && simTimers[i].time_end <= current_time)
		{
			//printf("(%2d) timer %d fired, start %lu, end %lu\n",station_id, i, simTimers[i].time_start, simTimers[i].time_end);
			//while(1);

			simTimers[i].isActive = 0;
			
			if(i == 6)
			{
				warpmac_pollUserIO();
				//warp_timer_start(USERIO_TIMER);	
			} else if (i == 7)
			{
				warp_timer_resetDone(PKTGEN_TIMER);
				//warpmac_incrementLEDLow();
				pktGen_allow = 1;

				warp_timer_setTimer(PKTGEN_TIMER, 0, dummyPacketInterval*TIMERCLK_CYCLES_PER_USEC);
				warp_timer_setMode(PKTGEN_TIMER, TIMER_MODE_CARRIERSENSE);
				warp_timer_start(PKTGEN_TIMER);
			} else
			{
				warp_timer_resetDone(i);
				usr_timerCallback(i);
			}

			if(current_time - simTimers[i].time_end > 100000 * SIM_SLOWDOWN)
			{
				printf("timer jitter at %ld : station: %d timer: %d diff:%ld\n", current_time, station_id, i,  current_time - simTimers[i].time_end);
			}
		}
	}

	if(csmaActive == 1 && /*simTimers[0].mode == TIMER_MODE_CARRIERSENSE
			&&*/ last_carrierSense == 1 && *carrier_sense == 0) //falling edge
	{
		warp_timer_start(0);

		Marker m;
		m.station = warpmac_getMyId();
		m.marker_type = MARKER_CSMA_FALLING_EDGE;
		m.time = get_time_nsec();
		m.added_to_list = 0;
		add_marker(&m, warpmac_getMyId());

		//printf("i");
	}

	last_carrierSense = *carrier_sense;

	warpmac_pollUserIO();
}

char warpmac_finishPhyRecv(){
	while(*station_rx_payload_indicator == RX_NO_PAYLOAD)
	{
		usleep(10);
	}

	if(*station_rx_payload_indicator == RX_GOOD_PAYLOAD)
	{
		return PHYRXSTATUS_PAYLOAD & PHYRXSTATUS_GOOD;
	}
	else if(*station_rx_payload_indicator == RX_BAD_PAYLOAD)
	{
		return PHYRXSTATUS_PAYLOAD & PHYRXSTATUS_BAD;
	}
	else
	{
		return 0;
	}
}

void warpmac_incrementLEDLow() {
	station_led_low = !station_led_low;
	//printf("Station %d Low LED %d\n", station_id, station_led_low);
}

void warpmac_incrementLEDHigh() {
	station_led_high = !station_led_high;
	//printf("Station %d High LED %d\n", station_id, station_led_high);
}

void warpmac_leftHex(unsigned char x) {
	station_left_hex = x;
	//printf("Station %d Left HEX %x\n", station_id, station_left_hex);
}

void warpmac_rightHex(unsigned char x) {
	station_right_hex = x;
	//printf("Station %d Right HEX %x\n", station_id, station_right_hex);
}

void warpmac_middleHex(unsigned char x) {
	station_middle_hex = x;
	//printf("Station %d Middle HEX %x\n", station_id, station_middle_hex);
}

void warpmac_setCSMA(char mode) {
	csmaActive = (mode == 0) ? 0 : 1;
}

//===================================================================================================================================================
//------NOT SCOPE OF SIMULATION-------
//===================================================================================================================================================

void warpmac_setBaseRate(unsigned char rate)  
{
	return;
}

int warpmac_setAntennaMode(unsigned int txMode, unsigned int rxMode)  
{
	return 0;
}

void warpmac_setDebugGPIO(unsigned char val, unsigned char mask)
{
	return;
}

void warpmac_waitForDMA()
{
	return;
}

//===================================================================================================================================================
//------UNCHANGED-------
//===================================================================================================================================================

///@brief Loads a header into the PHY, but does not immediately transmit
///
///This function performs the conversion from the packet structure to the byte array
///but does not send the packet over the air. This is used to "preload" and ACK
///into the PHY while the data packet is still being received. This extra
///pipelining saves on turn-around time.
///@param packet Pointer to user's Macframe
///@param buffer Packet buffer to send from
void warpmac_prepPhyForXmit(Macframe* packet, unsigned char buffer) {

	unsigned short origLength;
	if(packet->header.length == 0) {
		//All-header packet, like an ACK, has only base-rate symbols
		//Update the PHY packet length to consist of just the header
		origLength = 0;
		packet->header.length = NUM_HEADER_BYTES;
	}
	else {
		//Update the header's numBytes field to include the header, CRC and tail bytes
		origLength = packet->header.length;
		packet->header.length = packet->header.length + NUM_HEADER_BYTES + NUM_PAYLOAD_CRC_BYTES + NUM_PAYLOAD_TAIL_BYTES;
	}

	//Copy the header to the PHY packet buffer, filling exactly NUM_HEADER_BYTES bytes
	// The rest of the packet (its payload) will be copied by DMA later
	memcpy( (void *) warpphy_getBuffAddr(buffer), (void *) packet, (size_t) NUM_HEADER_BYTES);

	//Revert the packet payload length so the user code doesn't see the MAC/PHY tweaking
	packet->header.length = origLength;

	return;
}

///@brief Sends the current txBuffer's content.
///
///This function sends an existing Macframe over the air. This existing Macframe comes from the warpmac_prepPhyForXmit
///function.
void warpmac_startPhyXmit(unsigned char buffer) {

	int txResult;

	//Hold the packet detecter in reset to prevent any interfering receptions
	//	ofdm_pktDetector_mimo_WriteReg_pktDet_reset(PKTDET_BASEADDR, 1);

	//Transmit the packet; WARPPHY disables the radio Rx, enables Tx, sends the packet, then re-enables radio Rx
	// TXBLOCK will cause this call to block until the transmission is finished
	warpphy_setBuffs(buffer, controlStruct.pktBuf_phyRx);
	controlStruct.pktBuf_phyTx = buffer;

	txResult = warpphy_pktTx(TXNOBLOCK);

	if(txResult != 0)
	{
		xil_printf("WARPMAC tried to Tx, but PHY was busy\r\n");
	}

	return;
}

///@brief Blocks on PHY transmission
///
///This function waits for the PHY to finish transmitting, then re-enables wireless reception and sets packet buffers back
///to their default values.
inline void warpmac_finishPhyXmit() {
	warpphy_waitForTx();
}

signed char warpmac_setEMACQueue(unsigned short int priorities, unsigned short int topLength , unsigned char enOverwrite) {
	if (!eq_queue_empty(-1)) {
		xil_printf("warpmac_setEMACQueue: Make sure to empty the software ethernet queue prior to reinitializing it.\n");
		return -1;
	}

	if (eq_init(priorities, topLength, enOverwrite) != CALLBACK_OK) {
		xil_printf("eq_init: Error in initializing the software ethernet queue.\n");
		return -1;
	}
	controlStruct.emacQueue = 1;

	return 0;
}

///@brief Tells the PHY which piece of memory to receive to
///
///Also, it updates the global struct element Maccontrol#rxBuffIndex to keep track of that information*/
void warpmac_setRxBuffers(Macframe* rxFrame, unsigned char phyRxBuff) {

	//Update the gloal Rx Macframe
	rxPacket = rxFrame;

	//Update the PHY Rx buffer assignment
	controlStruct.pktBuf_phyRx = phyRxBuff;
//	warpphy_setBuffs(controlStruct.pktBuf_emacRx, controlStruct.pktBuf_phyRx);
	warpphy_setBuffs(controlStruct.pktBuf_phyTx, controlStruct.pktBuf_phyRx);

	return;
}

///@brief Tells the PHY which piece of memory to send from
void warpmac_setPHYTxBuffer(unsigned char txBuff) {
	controlStruct.pktBuf_phyTx = txBuff;
	warpphy_setBuffs(txBuff, controlStruct.pktBuf_phyRx);
	return;
}

///@brief Tells the PHY which piece of memory to send from
void warpmac_setEMACRxBuffer(unsigned char emacRxBuff) {
	controlStruct.pktBuf_emacRx = emacRxBuff;
	return;
}

unsigned char warpmac_getEMACRxBuffer() {
	return controlStruct.pktBuf_emacRx;
}

///@brief Enable the Ethernet
void warpmac_enableDataFromNetwork() {
	controlStruct.enableDataFromNetwork = 1;
	return;
}

///@brief Disables the Ethernet
void warpmac_disableDataFromNetwork() {
	controlStruct.enableDataFromNetwork = 0;
	return;
}

//===================================================================================================================================================
//------NOT IMPLEMENTED-------
//===================================================================================================================================================

void warpmac_setTimer(int type)
{
	unsigned int numSlots;
	int myRandNum;

	switch(type) {
		case TIMEOUT_TIMER:

			//User a 1-cycle slot, so the numSlots argument is equal to the number of clock cycles
			warp_timer_setTimer(TIMEOUT_TIMER, 0, controlStruct.timeout);
			warp_timer_setMode(TIMEOUT_TIMER, TIMER_MODE_NOCARRIERSENSE);
			warp_timer_start(TIMEOUT_TIMER);
			return;
			break;
		case BACKOFF_TIMER:
			myRandNum = randNum(controlStruct.currentCW);
			numSlots = myRandNum;
			if(controlStruct.currentCW < controlStruct.maxCW){
				controlStruct.currentCW = controlStruct.currentCW + 1;
			}
			warp_timer_setTimer(BACKOFF_TIMER, controlStruct.slotTime, numSlots);
			warp_timer_start(BACKOFF_TIMER);

			return;
			break;
		default:
			//Invalid timer ID; do nothing
			return;
			break;
	}
}

int warpmac_carrierSense() 
{
	usleep(controlStruct.slotTime);
	//printf("Medium is busy %d\n", *carrier_sense);
	if (*carrier_sense == CS_IDLE) {
		return 1;
	} else {
		return 0;
	}
}

void warpmac_setMaxResend(unsigned int c) 
{
	controlStruct.maxReSend = c;
}

void warpmac_setMaxCW(unsigned int c) 
{
	controlStruct.maxCW = c;
}

void warpmac_setTimeout(unsigned int time) 
{
	controlStruct.timeout = time * TIMERCLK_CYCLES_PER_USEC;
}

void warpmac_setSlotTime(unsigned int time) 
{
	controlStruct.slotTime = time * TIMERCLK_CYCLES_PER_USEC;
}

void warpmac_clearTimer(unsigned char theTimer) 
{
	warp_timer_pause(theTimer);
	warp_timer_resetDone(theTimer);

	return;
}

int warpmac_decrementRemainingReSend(Macframe* packet) 
{
	int status = 1;

	unsigned char remainingReSend;
	remainingReSend=(packet->header.remainingTx);

	if(remainingReSend>0){
		remainingReSend--;
	}

	else{
		status = 0;
	}

	packet->header.remainingTx = remainingReSend;

	return status;
}

int warpmac_inTimeout() 
{
	return 1;
}



void warpmac_resetCurrentCW() 
{
	controlStruct.currentCW = 0;
	return;
}



void warpmac_startTimer(unsigned char theTimer, unsigned char mode)
{
	printf("warpmac_startTimer(unsigned char theTimer, unsigned char mode) not implemented.\n");
}

unsigned int randNum(unsigned int N)
{
	return ((unsigned int)rand() >> (32-(N+5)));
}

int warpmac_inBackoff()
{
	printf("warpmac_inBackoff() not implemented.\n");
	return 0;
}



void warpmac_uartRecvHandler(void *CallBackRef)
{
	printf("warpmac_uartRecvHandler(void *CallBackRef) not implemented.\n");
}

inline void warpmac_setDummyPacketMode(char mode)
{
	if (mode) {
		xil_printf("Dummy Packet Mode Enabled\r\n");
		controlStruct.dummyPacketMode = 1;
	} else {
		xil_printf("Dummy Packet Mode Disabled\r\n");
		controlStruct.dummyPacketMode = 0;
	}
	return;
}

void warpmac_startPacketGeneration(unsigned int length, unsigned int interval)
{
	//printf("warpmac_startPacketGeneration(unsigned int length, unsigned int interval) not implemented.\n");
	startPktGeneration = 1;
	dummyPacketLength = length;
	dummyPacketInterval = interval;

	warp_timer_setTimer(PKTGEN_TIMER, 0, dummyPacketInterval*TIMERCLK_CYCLES_PER_USEC);
	warp_timer_setMode(PKTGEN_TIMER, TIMER_MODE_CARRIERSENSE); //ms
	warp_timer_start(PKTGEN_TIMER);
	return;
}

void warpmac_stopPacketGeneration()
{
	//printf("warpmac_stopPacketGeneration() not implemented.\n");
	startPktGeneration = 0;

	warp_timer_pause(PKTGEN_TIMER);
	warp_timer_resetDone(PKTGEN_TIMER);
}

void printBytes(unsigned char* data, int length)
{
	printf("printBytes(unsigned char* data, int length) not implemented.\n");
}

void warpmac_copyDataBetweenBuffers(unsigned char source, unsigned char destination, unsigned short length)
{
	printf("warpmac_copyDataBetweenBuffers(...) not implemented.\n");
}


