#ifndef WARPMAC_H
#define WARPMAC_H

#include "xparameters.h"
#include "warpphy.h"

//Define which radios get used below
//RADIOx_ADDR are defined by the radio controller driver
//Address of the first radio
#define FIRST_RADIO RADIO2_ADDR
//Address of the second radio
#define SECOND_RADIO RADIO3_ADDR

//Helper macros for constructing headers
//#define htons(A) ((((Xuint16)(A) & 0xff00) >> 8) | (((Xuint16)(A) & 0x00ff) << 8))
#define NODEID_TO_ADDR(theID)	( theID & 0xFFFF )
#define ADDR_TO_NODEID(theAddr)	( theAddr & 0xFFFF )

#define MY_XEM_MAX_FRAME_SIZE 2000

//Masks for supported modulation schemes; equivalent to number of bits per symbol
#define BPSK	1
#define QPSK	2
#define QAM_16	4
#define QAM_64	6

//Constants used for packet headers specifying full-rate modulation per-packet
// This constant fills in an 8-bit field in the header
// Each 4-bit nibble corresponds to an antenna
#define HDR_FULLRATE_BPSK		(BPSK | (BPSK<<4))
#define HDR_FULLRATE_QPSK		(QPSK | (QPSK<<4))
#define HDR_FULLRATE_QAM_16	(QAM_16 | (QAM_16<<4))
#define HDR_FULLRATE_QAM_64	(QAM_64 | (QAM_64<<4))

//2.4 GHz Band
#define GHZ_2 1
//5 GHz Band
#define GHZ_5 0

//Average RSSI threshold for carrier sensing
#define THRESH_CARRIER_SENSE 5000

//Event ID's for callback registration
#define EVENT_UPBUTTON			1
#define EVENT_LEFTBUTTON		2
#define EVENT_RIGHTBUTTON		3
#define EVENT_MIDDLEBUTTON		4
#define EVENT_TIMER				5
#define EVENT_DATAFROMNETWORK	6
#define EVENT_MGMTPKT			7
#define EVENT_PHYGOODHEADER		8
#define EVENT_PHYBADHEADER		9
#define EVENT_UARTRX			10

//The number of bytes in the header must be fixed and known by every node ahead of time
// It also must occupy an integral number of OFDM symbols (i.e. the base-rate symbols)
#define NUM_HEADER_BYTES 24

#define NUM_PAYLOAD_CRC_BYTES	4
//#define NUM_PAYLOAD_TAIL_BYTES	1 //Use for coded PHY
#define NUM_PAYLOAD_TAIL_BYTES	0	//Use for uncoded PHY

//Assign IDs for various timers
#define TIMEOUT_TIMER	0
#define BACKOFF_TIMER	1
#define USER_TIMER_A	2
#define USER_TIMER_B	3
#define USER_TIMER_C	4
#define USER_TIMER_D	5
#define USERIO_TIMER	6
#define PKTGEN_TIMER	7

//Enable polling of timers that are actually used (6-7 are always enabled)
#define POLL_TIMER0
#define POLL_TIMER1
#define POLL_TIMER2
#define POLL_TIMER3
#define POLL_TIMER4
#define POLL_TIMER5

//Number of clock cycles in 1us (X for XMHz PLB hosting the timer core)
#define TIMERCLK_CYCLES_PER_USEC 80
#define TIMERCLK_CYCLES_PER_MSEC 80000

//Poll the user I/O every 1 msec
#define USERIO_POLLRATE 1000*TIMERCLK_CYCLES_PER_USEC

///Structure contains the header contents of a packet
typedef struct {
	///Physical layer header struct
	phyHeader header;
} Macframe __attribute__((__aligned__(8)));

///Structure of miscellaneous control bits needed for correct operation of the MAC.
typedef struct {
	///Timeout time of system
	volatile unsigned int timeout;
	///Smallest time interval in system
	volatile unsigned int slotTime;
	///Maximum number of retransmissions before dropping a packet
	volatile unsigned char maxReSend;
	///Maximum contention window index: [0, 2^(maxCW+4)-1]
	volatile unsigned char maxCW;
	///Current contention window index
	volatile unsigned char currentCW;
	///Whether the Emac is currently enabled
	volatile unsigned char enableDataFromNetwork;
	volatile unsigned char pktBuf_phyTx;
	volatile unsigned char pktBuf_phyRx;
	volatile unsigned char pktBuf_emacRx;
	volatile unsigned char emacQueue;
	volatile unsigned char dummyPacketMode;

	///Constellation order of the full rate symbols
	volatile unsigned char mod_fullRateA;
	//volatile unsigned char mod_fullRateB;
} Maccontrol;

//Function prototypes

void warpmac_init();
void nullCallback(void* param);
int nullCallback_i(void* param);
void warpmac_pollPeripherals();
void warpmac_pollPhy();
void phyRx_goodHeader_handler();
void phyRx_badHeader_handler();
void warpmac_setCallback(int eventID, void(*handler)());
void warpmac_pollTimer();
void warpmac_pollDataSource();
void warpmac_incrementLEDLow();
void warpmac_incrementLEDHigh();
#define warpmac_incrementLEDredLow() (0)
#define warpmac_incrementLEDredHigh() (0)
void warpmac_leftHex(unsigned char x);
void warpmac_rightHex(unsigned char x);
void warpmac_middleHex(unsigned char x);
void warpmac_setTimer(int type);
void warpmac_prepPhyForXmit(Macframe* packet, unsigned char buffer);
void warpmac_startPhyXmit(unsigned char buffer);
void warpmac_finishPhyXmit();
void warpmac_enableDataFromNetwork();
void warpmac_disableDataFromNetwork();
void warpmac_setCSMA(char mode);








void emacRx_handler();
void warpmac_waitForDMA();
void warpmac_clearTimer(unsigned char theTimer);
void warpmac_startTimer(unsigned char theTimer, unsigned char mode);

unsigned int randNum(unsigned int N);
int warpmac_carrierSense();
void warpmac_prepPktToNetwork(void* thePkt, unsigned int length);
void warpmac_startPktToNetwork(unsigned int length);

void warpmac_setBaseRate(unsigned char rate);
int warpmac_setAntennaMode(unsigned int txMode, unsigned int rxMode);



void warpmac_setRxBuffers(Macframe* rxFrame, unsigned char phyRxBuff);
void warpmac_setPHYTxBuffer(unsigned char txBuff);
void warpmac_setEMACRxBuffer(unsigned char emacRxBuff);
int warpmac_decrementRemainingReSend(Macframe* packet);
void warpmac_resetCurrentCW();
void warpmac_setMaxResend(unsigned int c);
void warpmac_setMaxCW(unsigned int c);
void warpmac_setTimeout(unsigned int time);
void warpmac_setSlotTime(unsigned int time);
int warpmac_inTimeout();
int warpmac_inBackoff();

void warpmac_setDebugGPIO(unsigned char val, unsigned char mask);

char warpmac_finishPhyRecv();
void warpmac_uartRecvHandler(void *CallBackRef);
void warpmac_setDummyPacketMode(char mode);
void warpmac_startPacketGeneration(unsigned int length, unsigned int interval);
void warpmac_stopPacketGeneration();
int warpmac_getMyId();

void printBytes(unsigned char* data, int length);

void warpmac_copyDataBetweenBuffers(unsigned char source, unsigned char destination, unsigned short length);

#define warpmac_getEMACPayload(a) ((void *)warpphy_getBuffAddr((unsigned char)(a))+NUM_HEADER_BYTES)
signed char warpmac_setEMACQueue(unsigned short int priorities, unsigned short int topLength , unsigned char enOverwrite);

//**********START***********//
// FPGA Board Specific Code //
//**************************//
#ifdef WARP_FPGA_BOARD_V1_2
void userIO_handler(void *InstancePtr);

unsigned char sevenSegmentMap(unsigned char x);


#elif defined WARP_FPGA_BOARD_V2_2
void userIO_handler(void *InstancePtr);
void warpmac_incrementLEDLow();
void warpmac_incrementLEDHigh();
void warpmac_leftHex(unsigned char x);
void warpmac_rightHex(unsigned char x);
void warpmac_middleHex(unsigned char x);
#endif//ifdef WARP_FPGA_BOARD_V1_2

#endif
