/*
 * simulator.c
 *
 *  Created on: Aug 2, 2014
 *      Author: andreas
 */

#include "simulator.h"

long preamble_time;
long header_time;
long standard_payload_time;

typedef struct {
	unsigned char CH;						/// Channel
	unsigned char FB;						/// #GHZ_2, #GHZ_5
	unsigned char AM;						/// 2 - AntennaA, 3 - AntennaB, 5 - SelDiv
	unsigned char PW;						/// Power (in [0 0x3F])

	unsigned char BR;						/// Base rate, #BPSK, #QPSK
	unsigned char CR;						/// Code rate
	unsigned char FR;						/// Full rate, #BPSK, #QPSK, #16QAM
	unsigned char PFR;						/// Full rate in packet header

	unsigned char BW;						/// Bandwidth (in MHz)
	unsigned char ASC;						/// Available SCs
	unsigned char ESC;						/// Existing SCs
	unsigned char S;						/// Samples per OFDM symbol w/o CPE
	unsigned char TSN;						/// Training symbol #
	unsigned char TSL;						/// Training symbol length w/o CPE (in samples)
	unsigned short int PL;					/// Preamble length, no CPE (in samples)
	unsigned char DCL;						/// Data CRC size (in Byte)
	unsigned char HL;						/// Header length (in Byte)
	double CPE;								/// CP Extension
	double ACR[4];							/// Available inverted CR values

	double PT;								/// Processing time (in ms)
} simPhySetup;

#define TRMAC_CODERATE				HDR_CODE_RATE_12	/// code rate used for data packets
#define TRMAC_HEADERMOD				BPSK				/// header modulation
#define TRMAC_PACKETMOD				QPSK				/// packet modulation


simPhySetup sim_internals = {
		19,
		GHZ_5,
		5,
		0x3F,

		TRMAC_HEADERMOD,
		TRMAC_CODERATE,
		TRMAC_PACKETMOD,
		0,

		10,
		48,
		64,
		64,
		NUM_TRAINING_SYMBOLS,
		64,
		256,
		NUM_PAYLOAD_CRC_BYTES,
		NUM_HEADER_BYTES,
		1.25,
		{2, 1.5, 1.33333, 1},

#ifdef DEBUG
		0.05 /// 10-15us PHY latency, 5us for debug, 35us for SW turn-around (margin for code)
#else
		0.04 /// 10-15us PHY latency, 30us for SW turn-around (margin for code)
#endif
};

long get_preamble_time()
{
	unsigned int preamble = 0; // in samples

	preamble = sim_internals.PL + sim_internals.TSN * sim_internals.TSL;

	return (long)((preamble) * sim_internals.CPE / (double)sim_internals.BW / 1000. * 1000000L * SIM_SLOWDOWN); // x / 1e6 * 1e3 = 1e-3

}

long get_header_time()
{
	unsigned int preamble = 0, header = 0; // in samples

	preamble = sim_internals.PL + sim_internals.TSN * sim_internals.TSL;
	header = ceil(sim_internals.HL * 8. * sim_internals.ACR[ENABLE_CODING==0?3:0] / ((double)sim_internals.ASC * sim_internals.BR)) * sim_internals.S; // header CR always 1/2

	return (long)((preamble + header) * sim_internals.CPE / (double)sim_internals.BW / 1000. * 1000000L * SIM_SLOWDOWN); // x / 1e6 * 1e3 = 1e-3

}

long get_payload_time(int payload_size)
{
	unsigned int preamble = 0, header = 0, data = 0; // in samples

	preamble = sim_internals.PL + sim_internals.TSN * sim_internals.TSL;
	header = ceil(sim_internals.HL * 8. * sim_internals.ACR[ENABLE_CODING==0?3:0] / ((double)sim_internals.ASC * sim_internals.BR)) * sim_internals.S; // header CR always 1/2
	data = payload_size == 0 ? 0 : (ceil((payload_size+sim_internals.DCL) * 8. * sim_internals.ACR[ENABLE_CODING==0?3:sim_internals.CR] / ((double)sim_internals.ASC * sim_internals.FR)) * sim_internals.S);

	return (long)((preamble + header + data) * sim_internals.CPE / (double)sim_internals.BW / 1000. * 1000000L * SIM_SLOWDOWN); // x / 1e6 * 1e3 = 1e-3
}

void poll_stdin()
{
	char buf[10];
	if(read(STDIN_FILENO,buf,10) > 0)
	{
		if(buf[0] == 'q')
		{
			main_exit(0);
		}

		if(buf[0] == 's')
		{
			kill_stations();
		}

		if(buf[0] == 'l' || buf[0] == 'r' || buf[0] == 'u' || buf[0] == 'm' || buf[0] == 'p')
		{
			int s;

			if(sscanf(&(buf[1]), "%d", &s) == 1 && s >= 0 && s < SIM_STATION_COUNT)
			{
				if(buf[0] == 'l' && sim_left_buttons[s] == 0)
				{
					sim_left_buttons[s] = 1;
				}
				if(buf[0] == 'r' && sim_right_buttons[s] == 0)
				{
					sim_right_buttons[s] = 1;
				}
				if(buf[0] == 'u' && sim_up_buttons[s] == 0)
				{
					sim_up_buttons[s] = 1;
				}
				if(buf[0] == 'm' && sim_middle_buttons[s] == 0)
				{
					sim_middle_buttons[s] = 1;
				}
			}
			else
			{
				printf("Unknown Command. Station ID invalid.\n");
			}
		}
		else
		{
			printf("Unknown Command.\n");
		}
	}
}

void set_carrier_sense()
{
	int others_idle = 1;
	int j;
	for(j = 0; j < SIM_STATION_COUNT; ++j)
	{
		if(sim_tx_states[j] != TX_STATE_IDLE)
		{
			others_idle = 0;
		}
	}

	if(others_idle)
	{
		*carrier_sense = CS_IDLE;
	}
}

void poll_transmitters()
{
	//Look for active transmissions

	int i;
	for(i = 0; i < SIM_STATION_COUNT; ++i)
	{
		if(sim_tx_states[i] == TX_IDLE && sim_tx_active_indicators[i] == 1)
		{
			*carrier_sense = CS_BUSY;

			long current_time = get_time_nsec();

			sim_tx_states[i] = TX_ACTIVE;

			Macframe *frame = (Macframe*)((void*)sim_packet_buffers + SIM_PHY_BUFFER_COUNT * SIM_PHY_BUFFER_SIZE * i + SIM_PHY_BUFFER_SIZE * sim_tx_buffer_ids[i]);
			int length = frame->header.length;

			//printf("(%2d) tx packet: ", i);
			//print_packet_buffer(frame);

			station_current_history_items[i].station = i;
			station_current_history_items[i].frame = *frame;
			station_current_history_items[i].time_start = current_time;



			int j;
			for(j = 0; j < SIM_STATION_COUNT; ++j)
			{
				if(j != i)
				{
					sim_rx_payload_indicators[j] = RX_NO_PAYLOAD;
				}
			}


			struct event preamble_done_event;
			preamble_done_event.station = i;
			preamble_done_event.type = EVENT_PREAMBLE_DONE;
			preamble_done_event.time = current_time + preamble_time;

			station_current_history_items[i].time_header_start = preamble_done_event.time;

			enqueue_event(&preamble_done_event);

			if(length == NUM_HEADER_BYTES)
			{
				struct event header_payload_done_event;
				header_payload_done_event.station = i;
				header_payload_done_event.type = EVENT_HEADER_PAYLOAD_DONE;
				header_payload_done_event.time = current_time + header_time;

				station_current_history_items[i].time_payload_start = header_payload_done_event.time;
				station_current_history_items[i].time_end = header_payload_done_event.time;

				enqueue_event(&header_payload_done_event);
			}
			else
			{
				struct event header_done_event;
				header_done_event.station = i;
				header_done_event.type = EVENT_HEADER_DONE;
				header_done_event.time = current_time + header_time;
				enqueue_event(&header_done_event);

				station_current_history_items[i].time_payload_start = header_done_event.time;

				struct event payload_done_event;
				payload_done_event.station = i;
				payload_done_event.type = EVENT_PAYLOAD_DONE;

				if(length == NUM_HEADER_BYTES + SIM_STANDARD_PAYLOAD_SIZE)
				{
					payload_done_event.time = current_time + standard_payload_time;
				}
				else
				{
					payload_done_event.time = current_time + get_payload_time(length - NUM_HEADER_BYTES);
				}


				station_current_history_items[i].time_end = payload_done_event.time;

				enqueue_event(&payload_done_event);
			}



			for(j = 0; j < SIM_STATION_COUNT; ++j)
			{
				if(i != j)
				{
					int r = rand() % 1000000000;

					long len = station_current_history_items[i].time_end - station_current_history_items[i].time_start;

					if(r < 1000000000 / 100 * (100 - P_CORRUPTION))
					{
						long offset = (long)((double)r / (1000000000 / 100 * (100 - P_CORRUPTION)) * len);

						station_current_history_items[i].time_corruption[j] = station_current_history_items[i].time_start + offset;
					}
					else
					{
						station_current_history_items[i].time_corruption[j] = -1;
					}

				}
				else
				{
					station_current_history_items[i].time_corruption[j] = -1;
				}
			}


			int collision = 0;


			for(j=0; j<SIM_STATION_COUNT; j++)
			{
				if(j != i)
				{
					if(sim_tx_states[j] == TX_STATE_PREAMBLE)
					{
						sim_tx_states[j] = TX_STATE_PREAMBLE_CORRUPTED;
						station_current_history_items[j].time_collision = current_time;
					}

					if(sim_tx_states[j] == TX_STATE_HEADER)
					{
						sim_tx_states[j] = TX_STATE_HEADER_CORRUPTED;
						station_current_history_items[j].time_collision = current_time;
					}

					if(sim_tx_states[j] == TX_STATE_PAYLOAD)
					{
						sim_tx_states[j] = TX_STATE_PAYLOAD_CORRUPTED;
						station_current_history_items[j].time_collision = current_time;
					}

					if(sim_tx_states[j] != TX_IDLE)
					{
						collision = 1;
					}
				}
			}

			if(collision)
			{
				sim_tx_states[i] = TX_STATE_PREAMBLE_CORRUPTED;
				station_current_history_items[i].time_collision = current_time;
			}
			else
			{
				sim_tx_states[i] = TX_STATE_PREAMBLE;
				station_current_history_items[i].time_collision = -1;
			}


		}
	}
}

int rx_is_corrupted_preamble(int tx, int rx)
{
	if(station_current_history_items[tx].time_corruption[rx] >= station_current_history_items[tx].time_start && station_current_history_items[tx].time_corruption[rx] < station_current_history_items[tx].time_header_start)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

int rx_is_corrupted_header(int tx, int rx)
{
	if(station_current_history_items[tx].time_corruption[rx] >= station_current_history_items[tx].time_header_start && station_current_history_items[tx].time_corruption[rx] < station_current_history_items[tx].time_payload_start)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

int rx_is_corrupted_payload(int tx, int rx)
{
	if(station_current_history_items[tx].time_corruption[rx] >= station_current_history_items[tx].time_payload_start && station_current_history_items[tx].time_corruption[rx] < station_current_history_items[tx].time_end)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

void process_events()
{
	struct event next_event;

	int continue_processing = 1;

	while(continue_processing)
	{
		continue_processing = 0;

		if(peek_event(&next_event))
		{
			long current_time = get_time_nsec();

			if(next_event.time <= current_time)
			{
				dequeue_event(&next_event);

				//printf("processed at %lu event sta: %d, typ: %d, time: %lu, diff: %ld\n", current_time, next_event.station, next_event.type, next_event.time, next_event.time - current_time);
				continue_processing = 1;

				switch(next_event.type)
				{
				case EVENT_PREAMBLE_DONE:
					if(sim_tx_states[next_event.station] == TX_STATE_PREAMBLE)
					{
						//printf("(%2d) tx good preamble.\n", next_event.station);
						sim_tx_states[next_event.station] = TX_STATE_HEADER;
					}
					else if(sim_tx_states[next_event.station] == TX_STATE_PREAMBLE_CORRUPTED)
					{
						//printf("(%2d) tx bad preamble.\n", next_event.station);
					}
					else
					{
						printf("event processing error: EVENT_PREAMBLE_DONE vs tx_state %d\n", sim_tx_states[next_event.station]);
					}
					break;
				case EVENT_HEADER_DONE:
					if(sim_tx_states[next_event.station] == TX_STATE_HEADER)
					{
						//printf("(%2d) tx good header.\n", next_event.station);
						sim_tx_states[next_event.station] = TX_STATE_PAYLOAD;

						int j;
						for(j = 0; j < SIM_STATION_COUNT; ++j)
						{
							if(j != next_event.station)
							{
								if( rx_is_corrupted_preamble(next_event.station, j))
								{
									// do nothing
								}
								else if( rx_is_corrupted_header(next_event.station, j) )
								{
									sim_rx_header_indicators[j] = RX_BAD_HEADER;
								}
								else
								{
									memcpy(
											(void*)((void*)sim_packet_buffers + SIM_PHY_BUFFER_COUNT * SIM_PHY_BUFFER_SIZE * j + SIM_PHY_BUFFER_SIZE * sim_rx_buffer_ids[j]),
											(void*)((void*)sim_packet_buffers + SIM_PHY_BUFFER_COUNT * SIM_PHY_BUFFER_SIZE * next_event.station + SIM_PHY_BUFFER_SIZE * sim_tx_buffer_ids[next_event.station]),
										SIM_PHY_BUFFER_SIZE);

									sim_rx_header_indicators[j] = RX_GOOD_HEADER;
								}


							}
						}
					}
					else if(sim_tx_states[next_event.station] == TX_STATE_HEADER_CORRUPTED)
					{
						//printf("(%2d) tx bad header.\n", next_event.station);
						int j;
						for(j = 0; j < SIM_STATION_COUNT; ++j)
						{

							if(j != next_event.station)
							{
								if( rx_is_corrupted_preamble(next_event.station, j))
								{
									// do nothing
								}
								else
								{
									sim_rx_header_indicators[j] = RX_BAD_HEADER;
								}
							}
						}
					}
					else if(sim_tx_states[next_event.station] == TX_STATE_PREAMBLE_CORRUPTED)
					{

					}
					else
					{
						printf("event processing error: EVENT_HEADER_DONE vs tx_state %d\n", sim_tx_states[next_event.station]);
					}
					break;
				case EVENT_PAYLOAD_DONE:
					if(sim_tx_states[next_event.station] == TX_STATE_PAYLOAD)
					{
						//printf("(%2d) tx good payload.\n", next_event.station);

						int j;
						for(j = 0; j < SIM_STATION_COUNT; ++j)
						{
							if(j != next_event.station)
							{
								if( rx_is_corrupted_preamble(next_event.station, j))
								{
									// do nothing
								}
								else if( rx_is_corrupted_payload(next_event.station, j)  || rx_is_corrupted_header(next_event.station, j))
								{
									sim_rx_payload_indicators[j] = RX_BAD_PAYLOAD;
								}
								else
								{
									sim_rx_payload_indicators[j] = RX_GOOD_PAYLOAD;
								}
							}
						}



						sim_tx_active_indicators[next_event.station] = TX_IDLE;
						sim_tx_states[next_event.station] = TX_STATE_IDLE;

						set_carrier_sense();



						add_history_item(&station_current_history_items[next_event.station]);

					}
					else if(sim_tx_states[next_event.station] == TX_STATE_PAYLOAD_CORRUPTED
								|| sim_tx_states[next_event.station] == TX_STATE_HEADER_CORRUPTED)
					{
						//printf("(%2d) tx bad payload.\n", next_event.station);

						int j;
						for(j = 0; j < SIM_STATION_COUNT; ++j)
						{
							if(j != next_event.station)
							{
								if( rx_is_corrupted_preamble(next_event.station, j) )
								{
									// do nothing
								}
								else if( rx_is_corrupted_payload(next_event.station, j)  || rx_is_corrupted_header(next_event.station, j))
								{
									sim_rx_payload_indicators[j] = RX_BAD_PAYLOAD;
								}
								else
								{
									sim_rx_payload_indicators[j] = RX_BAD_PAYLOAD;
								}
							}
						}

						sim_tx_active_indicators[next_event.station] = TX_IDLE;
						sim_tx_states[next_event.station] = TX_STATE_IDLE;
						set_carrier_sense();
						add_history_item(&station_current_history_items[next_event.station]);
					}
					else if(sim_tx_states[next_event.station] == TX_STATE_PREAMBLE_CORRUPTED)
					{
						sim_tx_active_indicators[next_event.station] = TX_IDLE;
						sim_tx_states[next_event.station] = TX_STATE_IDLE;
						set_carrier_sense();
						add_history_item(&station_current_history_items[next_event.station]);
					}
					else
					{
						printf("event processing error: EVENT_PAYLOAD_DONE vs tx_state %d\n", sim_tx_states[next_event.station]);
					}
					break;
				case EVENT_HEADER_PAYLOAD_DONE:
					if(sim_tx_states[next_event.station] == TX_STATE_HEADER)
					{
						//printf("(%2d) tx good header & pl.\n", next_event.station);

						int j;
						for(j = 0; j < SIM_STATION_COUNT; ++j)
						{
							if(j != next_event.station)
							{
								if( rx_is_corrupted_preamble(next_event.station, j) )
								{
									// do nothing
								}
								else if( rx_is_corrupted_header(next_event.station, j) )
								{
									sim_rx_header_indicators[j] = RX_BAD_HEADER;
									sim_rx_payload_indicators[j] = RX_BAD_PAYLOAD;
								}
								else
								{
									memcpy(
											(void*)((void*)sim_packet_buffers + SIM_PHY_BUFFER_COUNT * SIM_PHY_BUFFER_SIZE * j + SIM_PHY_BUFFER_SIZE * sim_rx_buffer_ids[j]),
											(void*)((void*)sim_packet_buffers + SIM_PHY_BUFFER_COUNT * SIM_PHY_BUFFER_SIZE * next_event.station + SIM_PHY_BUFFER_SIZE * sim_tx_buffer_ids[next_event.station]),
										SIM_PHY_BUFFER_SIZE);

									sim_rx_header_indicators[j] = RX_GOOD_HEADER;
									sim_rx_payload_indicators[j] = RX_GOOD_PAYLOAD;
								}
							}
						}

						sim_tx_active_indicators[next_event.station] = TX_IDLE;
						sim_tx_states[next_event.station] = TX_STATE_IDLE;
						set_carrier_sense();
						add_history_item(&station_current_history_items[next_event.station]);
					}
					else if(sim_tx_states[next_event.station] == TX_STATE_HEADER_CORRUPTED)
					{
						int j;
						for(j = 0; j < SIM_STATION_COUNT; ++j)
						{
							if(j != next_event.station)
							{
								if( rx_is_corrupted_preamble(next_event.station, j) )
								{
									// do nothing
								}
								else
								{
									sim_rx_header_indicators[j] = RX_BAD_HEADER;
								}
							}
						}

						sim_tx_active_indicators[next_event.station] = TX_IDLE;
						sim_tx_states[next_event.station] = TX_STATE_IDLE;
						set_carrier_sense();
						add_history_item(&station_current_history_items[next_event.station]);
					}
					else if(sim_tx_states[next_event.station] == TX_STATE_PREAMBLE_CORRUPTED)
					{
						sim_tx_active_indicators[next_event.station] = TX_IDLE;
						sim_tx_states[next_event.station] = TX_STATE_IDLE;
						set_carrier_sense();
						add_history_item(&station_current_history_items[next_event.station]);
					}
					else
					{
						printf("event processing error: EVENT_HEADER_PAYLOAD_DONE vs tx_state %d\n", sim_tx_states[next_event.station]);
					}
					break;
				default:
					break;
				}
			}
		}
	}
}

int simulator_main()
{
	preamble_time = get_preamble_time();
	header_time = get_header_time();
	standard_payload_time = get_payload_time(SIM_STANDARD_PAYLOAD_SIZE);

	init_history();

	printf("packet times: %ld, %ld, %ld\n", preamble_time, header_time, standard_payload_time);

	//printf("SIM_INFO: Press key to start simulation.\n");
	//getchar();

	sleep(1);

	unlock_semaphore(SEM_SIM_START);

	long opts = fcntl(STDIN_FILENO, F_GETFL);
	fcntl(STDIN_FILENO, F_SETFL, opts | O_NONBLOCK);

	while(1)
	{
		poll_stdin();
		poll_transmitters();
		process_events();
	}

	return 0;
}
