/**
 * @file	eth_queue.h
 * @author	cd185990
 * @date	2010/10/07
 * @brief	PriorityQueue Header File
 *
 * Last Change Base: 		$LastChangedRevision: 456 $
 * Last Changed By: 		$LastChangedBy: cd185990 $
 * Last Changed On: 		$LastChangedDate: 2014-03-03 15:35:00 +0100 (Mon, 03 Mar 2014) $
 *
 * Copyright: 2010 - 2014, Mobile Network Performance Group, RWTH Aachen
 */

#ifndef ETHERNET_QUEUE_H_
#define ETHERNET_QUEUE_H_

#include "stdbool.h"


#define EQ_VERSION		"1.2.5"

/// Number of available BRAM buffers
#define EQ_MAX_LENGTH	32
/// Number of reserved buffers
#define EQ_RESERVED		13
/// IDs of reserved buffers
#define EQ_RESID_TB		31	/// temporary buffer
#define EQ_RESID_RX		30	/// PHY RX buffer
#define EQ_RESID_AR		29	/// autoresponse system buffer
#define EQ_RESID_TR		28	/// token ring buffer
#define EQ_RESID_ER		27	/// echoring buffer
#define EQ_RESID_DBG	26
#define EQ_RESID_MEAS2	25
#define EQ_RESID_MEAS1	19

typedef struct eq_result_t {
	unsigned char buffer;
	unsigned short int length;
} eq_result;

/**
 * @brief Initialize the software ethernet priority queue.
 *
 * @param priorities - Number of priorities beside top priority.
 * @param top_length - Length of the top priority queue. The other queues have the same size.
 * @param enable_overwrite - Enables overwriting values in the ring buffer if buffer is full.
 *
 * @return CALLBACK_OK | CALLBACK_ERROR
 */
signed char eq_init(unsigned char priorities, unsigned char top_length, bool enable_overwrite);

/**
 * @brief Creates an entry in the appropriate priority queue, and returns that entry ID.
 *
 * @param priority - Priority of the entry.
 * @param length - Length of the packet
 *
 * @param eq_result - Result of operation.
 * @return CALLBACK_OK | CALLBACK_ERROR
 */
signed char eq_enqueue(eq_result *result, unsigned char priority, unsigned short int length);

/**
 * @brief Returns the earliest inserted entry (with respect to its priority).
 *
 * @param eq_result - Result of operation.
 * @return CALLBACK_OK | CALLBACK_ERROR
 */
signed char eq_dequeue(eq_result *result);

/**
 * @brief Checks if the software ethernet priority queue of a certain priority is completely empty.
 *
 * @param priority - Priority to ask for, or -1 to check all priorities
 * @return true | false
 */
bool eq_queue_empty(signed char priority);

/**
 * @brief Checks if the software ethernet priority queue of a certain priority is completely full.
 *
 * @param priority - Priority to ask for, or -1 to check all priorities
 * @return true | false
 */
bool eq_queue_full(signed char priority);

/**
 * @brief Returns the information of the top most element in the queue without dequeueing it.
 *
 * @param eq_result - Result of operation.
 * @return CALLBACK_OK | CALLBACK_ERROR
 */
signed char eq_front(eq_result *result);


#ifndef CALLBACK_ERROR
	#define CALLBACK_ERROR -1
#endif
#ifndef CALLBACK_OK
	#define CALLBACK_OK 0
#endif
#ifndef even
	#define even( a ) (((a)%2 == 0) ? 1 : 0)
#endif

#endif /* ETHERNET_QUEUE_H_ */
